
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifndef E_M365_ENUMTYPES_H
#define E_M365_ENUMTYPES_H

#include <glib-object.h>

G_BEGIN_DECLS

/* Enumerations from "e-m365-enums.h" */
#define E_TYPE_M365_FOLDER_KIND	(e_m365_folder_kind_get_type())
GType e_m365_folder_kind_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_M365_STORE_OOO_ALERT_STATE	(camel_m365_store_ooo_alert_state_get_type())
GType camel_m365_store_ooo_alert_state_get_type	(void) G_GNUC_CONST;

G_END_DECLS

#endif /* E_M365_ENUMTYPES_H */

/* Generated data ends here */

