/*
 * SPDX-FileCopyrightText: (C) 2017 Red Hat, Inc. (www.redhat.com)
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#ifndef E_EWS_CONFIG_LOOKUP_H
#define E_EWS_CONFIG_LOOKUP_H

#include <glib-object.h>

G_BEGIN_DECLS

void e_ews_config_lookup_type_register (GTypeModule *type_module);

G_END_DECLS

#endif /* E_EWS_CONFIG_LOOKUP_H */
