
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "e-m365-enumtypes.h"
/* enumerations from "e-m365-enums.h" */
#include "e-m365-enums.h"
GType
e_m365_folder_kind_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GEnumValue values[] = {
			{ E_M365_FOLDER_KIND_UNKNOWN,
			  "E_M365_FOLDER_KIND_UNKNOWN",
			  "unknown" },
			{ E_M365_FOLDER_KIND_MAIL,
			  "E_M365_FOLDER_KIND_MAIL",
			  "mail" },
			{ E_M365_FOLDER_KIND_CALENDAR,
			  "E_M365_FOLDER_KIND_CALENDAR",
			  "calendar" },
			{ E_M365_FOLDER_KIND_CONTACTS,
			  "E_M365_FOLDER_KIND_CONTACTS",
			  "contacts" },
			{ E_M365_FOLDER_KIND_ORG_CONTACTS,
			  "E_M365_FOLDER_KIND_ORG_CONTACTS",
			  "org-contacts" },
			{ E_M365_FOLDER_KIND_USERS,
			  "E_M365_FOLDER_KIND_USERS",
			  "users" },
			{ E_M365_FOLDER_KIND_PEOPLE,
			  "E_M365_FOLDER_KIND_PEOPLE",
			  "people" },
			{ E_M365_FOLDER_KIND_SEARCH,
			  "E_M365_FOLDER_KIND_SEARCH",
			  "search" },
			{ E_M365_FOLDER_KIND_TASKS,
			  "E_M365_FOLDER_KIND_TASKS",
			  "tasks" },
			{ E_M365_FOLDER_KIND_MEMOS,
			  "E_M365_FOLDER_KIND_MEMOS",
			  "memos" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_enum_register_static (
			g_intern_static_string ("EM365FolderKind"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
camel_m365_store_ooo_alert_state_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GEnumValue values[] = {
			{ CAMEL_M365_STORE_OOO_ALERT_STATE_UNKNOWN,
			  "CAMEL_M365_STORE_OOO_ALERT_STATE_UNKNOWN",
			  "unknown" },
			{ CAMEL_M365_STORE_OOO_ALERT_STATE_NOTIFIED,
			  "CAMEL_M365_STORE_OOO_ALERT_STATE_NOTIFIED",
			  "notified" },
			{ CAMEL_M365_STORE_OOO_ALERT_STATE_CLOSED,
			  "CAMEL_M365_STORE_OOO_ALERT_STATE_CLOSED",
			  "closed" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_enum_register_static (
			g_intern_static_string ("CamelM365StoreOooAlertState"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}

/* Generated data ends here */

